package com.adaptrex.core.persistence.jpa;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.enterprise.context.ApplicationScoped;
import javax.inject.Inject;
import javax.persistence.EntityManager;
import javax.persistence.Query;
import javax.persistence.metamodel.EntityType;
import javax.persistence.metamodel.Metamodel;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.adaptrex.core.ext.data.Config;
import com.adaptrex.core.ext.data.Filter;
import com.adaptrex.core.ext.data.Sorter;
import com.adaptrex.core.ext.data.Store;
import com.adaptrex.core.persistence.AdaptrexPersistence;
import com.adaptrex.core.utilities.StringUtilities;

@ApplicationScoped
public class JpaPersistence implements AdaptrexPersistence {
	public JpaPersistence() {}
	
	private Map<String,Class<?>> entityCache = new HashMap<String,Class<?>> ();

	private Logger log = LoggerFactory.getLogger(JpaPersistence.class);
	@Inject private JpaPersistenceManager persistenceManager;

	private static SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd");
	
	@Override
	public Class<?> getEntityClass(String className) {
		return getEntityClass(className, null);
	}

	@Override
	public Class<?> getEntityClass(String className, String persistenceUnit) {
		try {
			if (entityCache.containsKey(className)) {
				return entityCache.get(className);
			}
			
			EntityManager em = persistenceUnit == null ? 
					getEntityManager() : getEntityManager(persistenceUnit);
			Metamodel meta = em.getMetamodel();
			
			for (EntityType<?> et : meta.getEntities()) {
				if (et.getJavaType().getName().endsWith("." + className)) {
					Class<?> clazz = et.getJavaType();
					entityCache.put(className, clazz);
					return clazz;
				}
			}
		} catch (Exception e) {
			log.warn("Error", e);
		}
		return null;
	}
	
	@Override
	public EntityManager getEntityManager() {
		try {
			return persistenceManager.getEntityManagerFactory().createEntityManager();
		} catch (Exception e) {
			log.warn("Error", e);
			return null;
		}
	}
	
	@Override
	public EntityManager getEntityManager(String puName) {
		try {
			return persistenceManager.getEntityManagerFactory().createEntityManager();
		} catch (Exception e) {
			log.warn("Error", e);
			return null;
		}
	}

	
	
	/* *******************************************************************
	 * 
	 * Get Single Entity
	 * 
	 */
	@Override
	public Object getEntity(String entityName, Integer id) {
		EntityManager em = getEntityManager();
		return em.find(getEntityClass(entityName), id);
	}

	
	@Override
	public Object getEntity(String entityName, Integer id, String persistenceUnitName) {
		// TODO: Need to implement this
		return null;
	}

	@Override
	public Object getEntity(Class<?> clazz, Integer id) {
		EntityManager em = getEntityManager();
		return em.find(clazz, id);
	}

	
	@Override
	public Object getEntity(Class<?> clazz, Integer id, String persistenceUnitName) {
		// TODO: Need to implement this
		return null;
	}
	
	
	/* *******************************************************************
	 * 
	 * Save Single Entity
	 * 
	 */
	@Override
	public Object saveEntity(Object entity) {
		EntityManager em = getEntityManager();
		em.getTransaction().begin();
		entity = em.merge(entity);
		em.flush();
		em.getTransaction().commit();
		return entity;
	}

	
	@Override
	public Object saveEntity(Object entity, String persistenceUnitName) {
		// TODO: Need to implement this
		return null;
	}	

	
	/* *******************************************************************
	 * 
	 * Delete Single Entity
	 * 
	 */
	@Override
	public Object deleteEntity(String entityName, Integer id) {
		EntityManager em = getEntityManager();
		Object entity= em.find(getEntityClass(entityName), id);
		em.getTransaction().begin();
		em.remove(entity);
		em.getTransaction().commit();
		return entity;
	}
	
	@Override
	public Object deleteEntity(String entityName, Integer id, String persistenceUnitName) {
		// TODO: Need to implement this
		return null;
	}	
	
	
	
	
	/* *******************************************************************
	 * 
	 * Entity List
	 * 
	 */
	@SuppressWarnings("unchecked")
	@Override
	public List<Object> getEntityList(Store store) {
		Config config = store.getConfig();
		List<Object> list = new ArrayList<Object>();
		
		String persistenceUnitName = config.getPersistenceUnitName();
		EntityManager em = persistenceUnitName == null ? 
				getEntityManager() : getEntityManager(persistenceUnitName);
		
		Map<String,Object> parameters = new HashMap<String,Object>(config.getParams());
				
		try {
			Class<?> clazz = config.getClazz();
			String className = clazz.getSimpleName();
			String alias = className.toLowerCase().substring(0,1);
			
			/*
			 * Core JPQL
			 */
			String jpql = "SELECT " + alias + " FROM " + className + " " + alias;
			
			/*
			 * Where
			 */ 
			String where = "";
			if (config.getWhere() != null) {
				where += config.getWhere();
			}
			
			/*
			 * Filter
			 */
			for (Filter filter : config.getFilters()) {
				if (!where.isEmpty()) where += " AND ";
				
				String val = String.valueOf(filter.getValue());
				String prop = filter.getProperty();
				
				
				/*
				 * Handle between inclusive
				 */
				if (val.contains(">=<")) {
					String[] filterParts = val.split(">=<"); 
					
					where += "(" + alias + "." + filter.getProperty() + " BETWEEN :" +
							prop + "_low AND :" + prop + "_high)" +
									"";
					parameters.put(prop + "_low", filterParts[0]);
					parameters.put(prop + "_high", filterParts[1]);
					
				/*
				 * Handle "OR"
				 */
				} else if (String.valueOf(filter.getValue()).contains("||")) {
					String whereOr = "";
					int i = 1;
					
					for (String orPart : String.valueOf(filter.getValue()).split("\\|\\|")) {
						whereOr += whereOr.isEmpty() ? "(" : " OR ";
						whereOr += alias + "." + filter.getProperty() + " = :" + filter.getProperty() + i;
						parameters.put(filter.getProperty() + i, orPart);
						i++;
					}
					whereOr += ")";
					where += whereOr;
					
				} else {
					String eqType = String.valueOf(filter.getValue()).contains("%") ? " like " : " = ";
					if (eqType.contains("like")) {
						where += "lower(" + alias + "." + filter.getProperty() + ")" + eqType + ":" + filter.getProperty();
					} else {
						where += alias + "." + filter.getProperty() + eqType + ":" + filter.getProperty();
					}
					
					parameters.put(filter.getProperty(), filter.getValue());
				}
			}
			
			
			/*
			 * Add where clause
			 */
			if (!where.isEmpty()) jpql += " WHERE " + where;
			
			
			if (config.getSorters().size() > 0) {
				String sortClause = "";
				for (Sorter sorter : config.getSorters()) {
					if (!sortClause.isEmpty()) sortClause += ",";
					sortClause += alias + "." + sorter.getProperty() + " " + sorter.getDirection(); 
				}
				jpql += " ORDER BY " + sortClause;
			}
			
			log.debug("JPQL: " + jpql);
			Query q = em.createQuery(jpql, clazz);
			
			/*
			 * Add params
			 * TODO: This is tedius... we want to automatically determine the
			 * type for the field we're trying to set the parameter for
			 */
			for (String key : parameters.keySet()) {
				Object paramObj = parameters.get(key);
				
				Class<?> paramType = q.getParameter(key).getParameterType();
				String typeName = paramType.getSimpleName().toLowerCase();
				
				try {
					if (typeName.equals("boolean")) {
						String stringVal = (String) paramObj;
						paramObj = stringVal.equals("1") || stringVal.equals("true") || stringVal.equals("Y");
						
					} else if (typeName.equals("date")) {
						paramObj = dateFormat.parse((String) paramObj);
						
					} else {
						Method m = paramType.getDeclaredMethod("valueOf", String.class);
						paramObj = m.invoke(null, paramObj);
					}
				} catch (Exception e) {}
				
				log.debug("  Param: " + key + " = " + paramObj);
				q.setParameter(key, paramObj);
			}

			store.setTotalCount(q.getResultList().size());
			
			if (config.getPageSize() != null) q.setMaxResults(config.getPageSize());
			if (config.getStart() != null) q.setFirstResult(config.getStart());
			list = q.getResultList();
		} catch (Exception e) {
			log.warn("Error", e);
		}
			
		em.close();
		return list;
	}

	
	/*
	 * The JPA implementation currently only handles a single persistence unit.  Return the
	 * entity list from the default persistence unit
	 */
	@Override
	public List<Object> getEntityList(Store store, String persistenceUnitName) {
		return getEntityList(store);
	}

	
	
	/* ************************************************************************************
	 * 
	 * Everything below this is experimental.  Possibly moved under a separate interface 
	 * 
	 */	
	@Override
	public boolean isIdField(Class<?> clazz, String fieldName) {
		return getMemberAnnotation(clazz, fieldName, "Id") != null;
	}

	@Override
	public boolean isIdField(Class<?> clazz, String fieldName,
			String persistenceUnitName) {
		return getMemberAnnotation(clazz, fieldName, "Id") != null;
	}
	
	/*
	 * TODO: Do we want to move ManyToMany to a separate method?
	 */
	@Override
	public boolean isOneToMany(Class<?> clazz, String fieldName) {
		return getMemberAnnotation(clazz, fieldName, "OneToMany") != null
				|| getMemberAnnotation(clazz, fieldName, "ManyToMany") != null;
	}

	@Override
	public boolean isOneToMany(Class<?> clazz, String fieldName,
			String persistenceUnitName) {
		return getMemberAnnotation(clazz, fieldName, "OneToMany") != null
				|| getMemberAnnotation(clazz, fieldName, "ManyToMany") != null;
	}

	@Override
	public boolean isManyToOne(Class<?> clazz, String fieldName) {
		return getMemberAnnotation(clazz, fieldName, "ManyToOne") != null;
	}

	@Override
	public boolean isManyToOne(Class<?> clazz, String fieldName,
			String persistenceUnitName) {
		return getMemberAnnotation(clazz, fieldName, "ManyToOne") != null;
	}	
	
	@Override
	public Object getFieldValue(Object entity, String fieldName) {
		Method getter = null;
		Class<?> clazz = entity.getClass();
		try {getter = clazz.getMethod("get" + capitalize(fieldName));} catch (Exception e) {};
		if (getter == null) {
			try {getter = clazz.getMethod("is" + capitalize(fieldName));} catch (Exception e) {};
		}
		if (getter == null) return null;
		
		try {
			return getter.invoke(entity);
		} catch (Exception e) {
			log.warn("Error", e);
			return null;
		}
	}

	@Override
	public Object getFieldValue(Object entity, String fieldName,
			String persistenceUnitName) {
		return getFieldValue(entity, fieldName);
	}
	
	
	
	
	
	/*
	 * This gets a list of all annotations on a class field (or it's getter)
	 */
	private List<Annotation> getMemberAnnotations(Class<?> clazz, String member) {
		Annotation[] annotations;
		try {
			Field field = clazz.getDeclaredField(StringUtilities.uncapitalize(member));
			if (field != null) {
				annotations = field.getDeclaredAnnotations();
				if (annotations.length > 0) {
					return Arrays.asList(annotations);
				}
			}
		} catch (Exception e) {}
		
		try {
			Method getter = clazz.getMethod("get" + member);
			if (getter == null) getter = clazz.getMethod("is" + member);
			if (getter != null) {
				annotations = getter.getDeclaredAnnotations();
				if (annotations.length > 0) {
					return Arrays.asList(annotations);
				}
			}
		} catch (Exception e) {}
		
		return new ArrayList<Annotation>();
	}
	
	/*
	 * This returns a specific annotation on a class field (or it's getter)
	 */
	private Annotation getMemberAnnotation(Class<?> clazz, String member, String annotationName) {
		for (Annotation annotation : getMemberAnnotations(clazz, member)) {
			if (annotation.annotationType().getSimpleName().equals(annotationName)) {
				return annotation;
			}
		}
		
		return null;
	}
	
	
	private static String capitalize(String string) {
		return string.substring(0,1).toUpperCase() + string.substring(1);
	}
}
